"""Otis binary output."""
from pathlib import Path
import numpy as np
import xarray as xr

from oceantide.core.otis import u_from_z, v_from_z, indices_open_boundary
from oceantide.core.utils import set_attributes


def to_otis_netcdf(self, dirname, hfile=True, ufile=True, gfile=True, suffix=None):
    """Write dataset as Otis binary format.

    Args:
        - self (oceantide.tide.Tide): Oceantide Tide instance.
        - dirname (str): Directory to save binary files.
        - hfile (bool): Save tidal elevation binary file.
        - ufile (bool): Save tidal transports binary file.
        - gfile (bool): Save grid file.
        - suffix (str): Suffix to define file names, by default defined by cons names.

    Return:
        - filename (dict): Name of files written.

    """
    ds = self._obj.transpose("con", "lon", "lat")
    ds = ds.rename({"con": "nc", "lon": "nx", "lat": "ny"})
    ds = ds.fillna(0.0)

    if suffix is None:
        suffix = "".join(list(ds.nc.values)).lower()

    lat_z, lon_z = np.meshgrid(ds.ny, ds.nx)
    dx = float(ds.nx[1] - ds.nx[0])
    dy = float(ds.ny[1] - ds.ny[0])
    ds = ds.drop(["nc", "nx", "ny"])
    cons = np.array([f"{c:4s}".lower() for c in self._obj.con.values], dtype="S4")

    mz = xr.where(ds.dep > 0, 1.0, 0.0)
    if ufile or gfile:
        mu = mz * mz.roll(nx=1, roll_coords=False)
        mv = mz * mz.roll(ny=1, roll_coords=False)
        lon_u = lon_z - dx
        lat_u = lat_z
        lon_v = lon_z
        lat_v = lat_z - dy

    filenames = {}

    # Write elevations
    if hfile:
        fname = f"h_{suffix.lstrip('_')}.nc" if suffix else "h.nc"
        filenames["hfile"] = Path(dirname) / fname

        ha = self.amplitude("h").transpose("con", "lon", "lat")
        ha = ha.rename({"con": "nc", "lon": "nx", "lat": "ny"}).drop(["nc", "nx", "ny"])
        hp = self.phase("h").transpose("con", "lon", "lat")
        hp = hp.rename({"con": "nc", "lon": "nx", "lat": "ny"}).drop(["nc", "nx", "ny"])
        dsout = xr.Dataset()
        dsout["con"] = xr.DataArray(cons, dims=("nc"))
        dsout["lon_z"] = xr.DataArray(lon_z, dims=("nx", "ny"))
        dsout["lat_z"] = xr.DataArray(lat_z, dims=("nx", "ny"))
        dsout["ha"] = xr.DataArray(ha, dims=("nc", "nx", "ny"))
        dsout["hp"] = xr.DataArray(hp, dims=("nc", "nx", "ny"))
        dsout["hRe"] = xr.DataArray(ds.h.real, dims=("nc", "nx", "ny"))
        dsout["hIm"] = xr.DataArray(ds.h.imag, dims=("nc", "nx", "ny"))

        set_attributes(dsout, "otis")
        dsout.attrs = {"title": "OTIS tidal elevation file", "source": "Oceantide"}

        dsout.to_netcdf(filenames["hfile"])

    # Write transports
    if ufile:
        fname = f"u_{suffix.lstrip('_')}.nc" if suffix else "u.nc"
        filenames["ufile"] = Path(dirname) / fname

        ua = self.amplitude("u").transpose("con", "lon", "lat")
        ua = ua.rename({"con": "nc", "lon": "nx", "lat": "ny"}).drop(["nc", "nx", "ny"])
        up = self.phase("u").transpose("con", "lon", "lat")
        up = up.rename({"con": "nc", "lon": "nx", "lat": "ny"}).drop(["nc", "nx", "ny"])
        va = self.amplitude("v").transpose("con", "lon", "lat")
        va = va.rename({"con": "nc", "lon": "nx", "lat": "ny"}).drop(["nc", "nx", "ny"])
        vp = self.phase("v").transpose("con", "lon", "lat")
        vp = vp.rename({"con": "nc", "lon": "nx", "lat": "ny"}).drop(["nc", "nx", "ny"])
        dsout = xr.Dataset()
        dsout["con"] = dsout["con"] = xr.DataArray(cons, dims=("nc"))
        dsout["lon_u"] = xr.DataArray(lon_u, dims=("nx", "ny"))
        dsout["lat_u"] = xr.DataArray(lat_u, dims=("nx", "ny"))
        dsout["lon_v"] = xr.DataArray(lon_v, dims=("nx", "ny"))
        dsout["lat_v"] = xr.DataArray(lat_v, dims=("nx", "ny"))
        dsout["ua"] = u_from_z(xr.DataArray(ua, dims=("nc", "nx", "ny")), mz, mu)
        dsout["up"] = u_from_z(xr.DataArray(up, dims=("nc", "nx", "ny")), mz, mu)
        dsout["va"] = v_from_z(xr.DataArray(va, dims=("nc", "nx", "ny")), mz, mv)
        dsout["vp"] = v_from_z(xr.DataArray(vp, dims=("nc", "nx", "ny")), mz, mv)
        dsout["URe"] = u_from_z(
            xr.DataArray(ds.u.real * ds.dep, dims=("nc", "nx", "ny")), mz, mu
        )
        dsout["UIm"] = u_from_z(
            xr.DataArray(ds.u.imag * ds.dep, dims=("nc", "nx", "ny")), mz, mu
        )
        dsout["VRe"] = v_from_z(
            xr.DataArray(ds.v.real * ds.dep, dims=("nc", "nx", "ny")), mz, mv
        )
        dsout["VIm"] = v_from_z(
            xr.DataArray(ds.v.imag * ds.dep, dims=("nc", "nx", "ny")), mz, mv
        )

        set_attributes(dsout, "otis")
        dsout.attrs = {
            "title": "OTIS tidal transport/current file",
            "source": "Oceantide",
        }

        dsout.to_netcdf(filenames["ufile"])

    # Write grid
    if gfile:
        fname = f"grid_{suffix.lstrip('_')}.nc" if suffix else "grid.nc"
        filenames["gfile"] = Path(dirname) / fname

        dsout = xr.Dataset()
        dsout["lon_z"] = xr.DataArray(lon_z, dims=("nx", "ny"))
        dsout["lat_z"] = xr.DataArray(lat_z, dims=("nx", "ny"))
        dsout["lon_u"] = xr.DataArray(lon_u, dims=("nx", "ny"))
        dsout["lat_u"] = xr.DataArray(lat_u, dims=("nx", "ny"))
        dsout["lon_v"] = xr.DataArray(lon_v, dims=("nx", "ny"))
        dsout["lat_v"] = xr.DataArray(lat_v, dims=("nx", "ny"))
        dsout["mz"] = xr.DataArray(mz, dims=("nx", "ny"))
        dsout["mu"] = xr.DataArray(mu, dims=("nx", "ny"))
        dsout["mv"] = xr.DataArray(mv, dims=("nx", "ny"))
        dsout["hz"] = xr.DataArray(ds.dep, dims=("nx", "ny"))
        dsout["hu"] = u_from_z(xr.DataArray(ds.dep, dims=("nx", "ny")), mz, mu)
        dsout["hv"] = v_from_z(xr.DataArray(ds.dep, dims=("nx", "ny")), mz, mv)
        dsout["iob_z"] = xr.DataArray(
            indices_open_boundary(dsout.mz), dims=("iiob", "nob_z")
        )
        dsout["iob_u"] = xr.DataArray(
            indices_open_boundary(dsout.mu), dims=("iiob", "nob_u")
        )
        dsout["iob_v"] = xr.DataArray(
            indices_open_boundary(dsout.mv), dims=("iiob", "nob_v")
        )

        set_attributes(dsout, "otis")
        dsout.attrs = {"title": "OTIS Arakawa C-grid file", "source": "Oceantide"}

        dsout.to_netcdf(filenames["gfile"])

    fname = f"model_{suffix.lstrip('_')}" if suffix else "model"
    with open(Path(dirname) / fname, mode="w") as stream:
        for filename in filenames.values():
            stream.write(f"./{filename.name}\n")

    return filenames


if __name__ == "__main__":
    from oceantide import read_otis_netcdf

    dset = read_otis_netcdf("/data/tide/tpxo9v4a/netcdf/DATA/Model_tpxo9v4a")
    dset = dset.isel(lon=slice(None, None, 10), lat=slice(None, None, 10)).load()
    filenames = dset.tide.to_otis_netcdf("./", hfile=True, ufile=True, gfile=True)
    print(filenames)



"""Top-level package for oceantide."""

__version__ = "0.4.0"
__author__ = "Oceanum Developers"
__contact__ = "developers@oceanum.science"
__url__ = "http://github.com/wavespectra/oceantide"
__description__ = "Library for ocean tide prediction"
__keywords__ = "ocean tide prediction constituents xarray accessor"


def _import_read_functions(pkgname="input"):
    """Make read functions available at module level.

    Functions are imported here if:
        they are defined in a module wavespectra.input.{modname}
        they are named as read_{modname}

    """
    import os
    import glob
    from importlib import import_module

    here = os.path.dirname(os.path.abspath(__file__))
    for filename in glob.glob1(os.path.join(here, pkgname), "*.py"):
        module = os.path.splitext(filename)[0]
        if module == "__init__":
            continue
        func_name = f"read_{module}"
        try:
            globals()[func_name] = getattr(
                import_module(f"oceantide.{pkgname}.{module}"), func_name
            )
        except Exception as exc:
            print(f"Cannot import reading function {func_name}:\n{exc}")


_import_read_functions()



"""Read Otis binary file format."""
import os
import xarray as xr

from oceantide.core.otis import (
    read_otis_bin_grid,
    read_otis_bin_h,
    read_otis_bin_u,
    otis_to_oceantide,
    otis_filenames,
)
from oceantide.tide import Tide


def read_otis_binary(filename=None, gfile=None, hfile=None, ufile=None):
    """Read Otis Binary file format.

    Args:
        - filename (str): Name of `Model_*` metadata file specifying other files to read.
        - gfile (str): Name of grid file to read, by default defined from `filename`.
        - hfile (str): Name of elevation file to read, by default defined from `filename`.
        - ufile (str): Name of currents file to read, by default defined from `filename`.

    Returns:
        - dset (Oceantide Dataset): Formatted dataset with the tide accessor.

    Note:
        - Otis data are usually provided in 4 separate files:
          - Meta file named as `Model_*` specifyig the 3 data files to read.
          - Grid file named as `grid*` with grid information and model depths.
          - Elevation file named as `h*` with elevation constituents data.
          - Transport file named as `uv` with transport constituents data.
          The path of the three data files can be prescribed either by specifying the
          meta file path (`filename`) or by explicitly providing their path
          (`gfile`, `hfile` and `ufile`).

    """
    if filename is not None:
        _gfile, _hfile, _ufile = otis_filenames(filename)
    else:
        _gfile = _hfile = _ufile = None
        if not all([gfile, hfile, ufile]):
            raise ValueError(
                "Either specify `filename` or all of `gfile`, `hfile`, `ufile`."
            )

    gfile = gfile or _gfile
    hfile = hfile or _hfile
    ufile = ufile or _ufile

    dsg = read_otis_bin_grid(gfile)
    dsh = read_otis_bin_h(hfile)
    dsu = read_otis_bin_u(ufile)

    dset = otis_to_oceantide(dsg, dsh, dsu)

    return dset


if __name__ == "__main__":

    hfile = "/data/tide/tpxo9v4a/bin/DATA/h_tpxo9.v4a"
    ufile = "/data/tide/tpxo9v4a/bin/DATA/u_tpxo9.v4a"
    gfile = "/data/tide/tpxo9v4a/bin/DATA/grid_tpxo9.v4a"

    dset = read_otis_binary(gfile=gfile, hfile=hfile, ufile=ufile)



"""Otis binary output."""
from pathlib import Path
import numpy as np

from oceantide.core.otis import (
    write_otis_bin_h,
    write_otis_bin_u,
    write_otis_bin_grid,
    u_from_z,
    v_from_z,
)


def to_otis_binary(self, dirname, hfile=True, ufile=True, gfile=True, suffix=None):
    """Write dataset as Otis binary format.

    Args:
        - self (oceantide.tide.Tide): Oceantide Tide instance.
        - dirname (str): Directory to save binary files.
        - hfile (bool): Save tidal elevation binary file.
        - ufile (bool): Save tidal transports binary file.
        - gfile (bool): Save grid file.
        - suffix (str): Suffix to define file names, by default defined by cons names.

    Return:
        - filename (dict): Name of files written.

    """
    ds = self._obj.transpose("con", "lon", "lat")
    ds = ds.rename({"con": "nc", "lon": "nx", "lat": "ny"})
    ds = ds.fillna(0.0)

    if suffix is None:
        suffix = "".join(list(ds.nc.values)).lower()

    filenames = {}

    # Write elevations
    if hfile:
        fname = f"h_{suffix.lstrip('_')}" if suffix else "h"
        filenames["hfile"] = Path(dirname) / fname
        write_otis_bin_h(
            hfile=filenames["hfile"],
            hRe=ds.h.real,
            hIm=ds.h.imag,
            con=ds.nc,
            lon=ds.nx,
            lat=ds.ny,
        )

    # Write transports
    if ufile:
        fname = f"u_{suffix.lstrip('_')}" if suffix else "u"
        filenames["ufile"] = Path(dirname) / fname
        mz = ds.dep > 0
        mu = mz * mz.roll(nx=1, roll_coords=False)
        mv = mz * mz.roll(ny=1, roll_coords=False)
        write_otis_bin_u(
            ufile=filenames["ufile"],
            URe=u_from_z(ds.u.real * ds.dep, mz, mu),
            UIm=u_from_z(ds.u.imag * ds.dep, mz, mu),
            VRe=v_from_z(ds.v.real * ds.dep, mz, mv),
            VIm=v_from_z(ds.v.imag * ds.dep, mz, mv),
            con=ds.nc,
            lon=ds.nx,
            lat=ds.ny,
        )

    # Write grid
    if gfile:
        fname = f"grid_{suffix.lstrip('_')}" if suffix else "grid"
        filenames["gfile"] = Path(dirname) / fname
        write_otis_bin_grid(
            gfile=filenames["gfile"],
            hz=ds.dep,
            mz=mz,
            lon=ds.nx,
            lat=ds.ny,
            dt=12.0,
        )

    fname = f"model_{suffix.lstrip('_')}" if suffix else "model"
    with open(Path(dirname) / fname, mode="w") as stream:
        for filename in filenames.values():
            stream.write(f"./{filename.name}\n")

    return filenames



"""Read Otis netcdf file format."""
import os
import xarray as xr

from oceantide.core.otis import otis_to_oceantide, otis_filenames
from oceantide.tide import Tide


def read_otis_netcdf(filename=None, gfile=None, hfile=None, ufile=None):
    """Read Otis Netcdf file format.

    Args:
        - filename (str): Name of `Model_*` metadata file specifying other files to read.
        - gfile (str): Name of grid file to read, by default defined from `filename`.
        - hfile (str): Name of elevation file to read, by default defined from `filename`.
        - ufile (str): Name of currents file to read, by default defined from `filename`.

    Returns:
        - dset (Oceantide Dataset): Formatted dataset with the tide accessor.

    Note:
        - Otis data are usually provided in 4 separate files:
          - Meta file named as `Model_*` specifyig the 3 data files to read.
          - Grid file named as `grid*.nc` with grid information and model depths.
          - Elevation file named as `h*.*.nc` with elevation constituents data.
          - Transport file named as `uv.*.nc` with transport constituents data.
          The path of the three data files can be prescribed either by specifying the
          meta file path (`filename`) or by explicitly providing their path
          (`gfile`, `hfile` and `ufile`).

    """
    if filename is not None:
        _gfile, _hfile, _ufile = otis_filenames(filename)
    else:
        _gfile = _hfile = _ufile = None
        if not all([gfile, hfile, ufile]):
            raise ValueError(
                "Either specify `filename` or all of `gfile`, `hfile`, `ufile`."
            )

    gfile = gfile or _gfile
    hfile = hfile or _hfile
    ufile = ufile or _ufile

    dsg = xr.open_dataset(gfile, chunks={})
    dsh = xr.open_dataset(hfile, chunks={})
    dsu = xr.open_dataset(ufile, chunks={})

    dset = otis_to_oceantide(dsg, dsh, dsu)

    return dset



"""The setup script."""
from setuptools import setup, find_packages

import oceantide


with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "click",
    "dask",
    "gcsfs",
    "netCDF4",
    "numpy",
    "scipy",
    "xarray",
    "zarr",
]

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest",
]

setup(
    author=oceantide.__author__,
    author_email=oceantide.__contact__,
    description=oceantide.__description__,
    keywords=oceantide.__keywords__,
    version=oceantide.__version__,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    entry_points={
        "console_scripts": [
            "oceantide=oceantide.cli:main",
        ],
    },
    install_requires=requirements,
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    name="oceantide",
    packages=find_packages(),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://gitlab.com/oceanum/oceantide",
    zip_safe=False,
)



"""Core tools for tidal analysis and prediction."""
from pathlib import Path
import yaml
import numpy as np
import xarray as xr

from oceantide.constituents import V0U


HERE = Path(__file__).parent


def nodal(time, con):
    """Nodal correction.

    Args:
        - time (1darray): Time given as the number of days since 01 Jan 1992 + 48622,
          or equivalently the number of days since 17 Nov 1858.
        - con (1darray): Constituents to compute.

    Returns:
        - pu (1darray): Nodal correction pu.
        - pf (1darray): Nodal correction pf.
        - v0u (1darray): Nodal correction v0u.

    """
    rad = np.pi / 180.0
    s, h, p, omega = astrol(time)
    sinn = np.sin(omega * rad)
    cosn = np.cos(omega * rad)
    sin2n = np.sin(2 * omega * rad)
    cos2n = np.cos(2 * omega * rad)
    sin3n = np.sin(3 * omega * rad)

    ndict = {
        "M2": {
            "f": np.sqrt(
                (1.0 - 0.03731 * cosn + 0.00052 * cos2n) ** 2
                + (0.03731 * sinn - 0.00052 * sin2n) ** 2
            ),
            "u": np.arctan(
                (-0.03731 * sinn + 0.00052 * sin2n)
                / (1.0 - 0.03731 * cosn + 0.00052 * cos2n)
            )
            / rad,
        },
        "S2": {"f": 1.0, "u": 0.0},
        "K1": {
            "f": np.sqrt(
                (1.0 + 0.1158 * cosn - 0.0029 * cos2n) ** 2
                + (0.01554 * sinn - 0.0029 * sin2n) ** 2
            ),
            "u": np.arctan(
                (-0.1554 * sinn + 0.0029 * sin2n)
                / (1.0 + 0.1158 * cosn - 0.0029 * cos2n)
            )
            / rad,
        },
        "O1": {
            "f": np.sqrt(
                (1.0 + 0.189 * cosn - 0.0058 * cos2n) ** 2
                + (0.189 * sinn - 0.0058 * sin2n) ** 2
            ),
            "u": 10.8 * sinn - 1.3 * sin2n + 0.2 * sin3n,
        },
        "N2": {
            "f": np.sqrt(
                (1.0 - 0.03731 * cosn + 0.00052 * cos2n) ** 2
                + (0.03731 * sinn - 0.00052 * sin2n) ** 2
            ),
            "u": np.arctan(
                (-0.03731 * sinn + 0.00052 * sin2n)
                / (1.0 - 0.03731 * cosn + 0.00052 * cos2n)
            )
            / rad,
        },
        "P1": {"f": 1.0, "u": 0.0},
        "K2": {
            "f": np.sqrt(
                (1.0 + 0.2852 * cosn + 0.0324 * cos2n) ** 2
                + (0.3108 * sinn + 0.0324 * sin2n) ** 2
            ),
            "u": np.arctan(
                -(0.3108 * sinn + 0.0324 * sin2n)
                / (1.0 + 0.2852 * cosn + 0.0324 * cos2n)
            )
            / rad,
        },
        "Q1": {
            "f": np.sqrt((1.0 + 0.188 * cosn) ** 2 + (0.188 * sinn) ** 2),
            "u": np.arctan(0.189 * sinn / (1.0 + 0.189 * cosn)) / rad,
        },
    }

    ncon = len(con)
    pu = np.zeros(ncon)
    pf = np.ones(ncon)
    v0u = np.zeros(ncon)

    for ii, vv in enumerate(con):
        if vv in ndict:
            pu[ii] = ndict[vv]["u"] * rad
            pf[ii] = ndict[vv]["f"]
        if vv in V0U.keys():
            v0u[ii] = V0U[vv]

    return pu, pf, v0u


def astrol(time):
    """Mean astronomical longitudes  s, h, p, N.

    Args:
        - time (1darray): Time given as the number of days since 01 Jan 1992 + 48622,
          or equivalently the number of days since 17 Nov 1858.

    Returns:
        - s (1darray): The mean longitude of moon.
        - h (1darray): The mean longitude of sun.
        - p (1darray): The mean longitude of lunar perigee.
        - N (1darray): The mean longitude of ascending lunar node.

    """
    T = time - 51544.4993
    s = (218.3164 + 13.17639648 * T) % 360
    h = (280.4661 + 0.98564736 * T) % 360
    p = (83.3535 + 0.11140353 * T) % 360
    N = (125.0445 - 0.05295377 * T) % 360
    return s, h, p, N


def arakawa_grid(nx, ny, x0, x1, y0, y1, variable):
    """Arakawa grid coordinates for variable.

    Args:
        - nx (int): Number of grid point along the x direction.
        - ny (int): Number of grid point along the y direction.
        - x0 (float): Left limit (degrees).
        - x1 (float): Right limit (degrees).
        - y0 (float): Bottom limit (degrees).
        - y1 (float): Left limit (degrees).
        - variable (str): Model variable, one of 'h', 'u', 'v'.

    Returns:
        - lon (1darray): Longitude coordinates in Arakawa grid for variable.
        - lat (1darray): Latitude coordinates in Arakawa for variable.

    """
    dx = (x1 - x0) / nx
    dy = (y1 - y0) / ny
    xcorner = np.clip(np.arange(x0, x1, dx), x0, x1)[0:nx]
    ycorner = np.clip(np.arange(y0, y1, dy), y0, y1)[0:ny]
    if variable == "h":
        lon = xcorner + dx / 2
        lat = ycorner + dy / 2
    elif variable == "u":
        lon = xcorner
        lat = ycorner + dy / 2
    elif variable == "v":
        lon = xcorner + dx / 2
        lat = ycorner
    else:
        raise ValueError(f"'variable' must be one of 'h', 'u', 'v', got {variable}")
    return lon, lat


def set_attributes(dset, dataset_type):
    """Set variable attributes for variables from given dataset type.

    Args:
        dset (Dataset): Dataset to set attributes from.docs
        dataset_type (str): Dataset type, e.g., 'otis'.

    """
    all_attrs = yaml.load(open(HERE / "attributes.yml"), Loader=yaml.Loader)
    if isinstance(dset, xr.Dataset):
        for varname, darr in dset.variables.items():
            darr.attrs.update(all_attrs.get(dataset_type).get(varname, {}))
    elif isinstance(dset, xr.DataArray):
        dset.attrs.update(all_attrs.get(dataset_type).get(dset.name, {}))


def compute_scale_and_offset(vmin, vmax, nbit=16):
    """Returns the scale_factor and add_offset for packing data.

    Args:
        - vmin (float): The minumum value in the array to pack.
        - vmax (float): The maximum value in the array to pack.
        - nbit (int): The number of bits into which you wish to pack.

    Returns:
        - scale_factor (float): The scale factor value to use for packing array.
        - add_offset (float): The add_offset value to use for packing array.

    """
    if vmax == vmin:
        scale_factor = 1
    else:
        scale_factor = (vmax - vmin) / (2 ** nbit - 1)
    add_offset = vmin + 2 ** (nbit - 1) * scale_factor
    return scale_factor, add_offset



"""Read oceantide file format."""
import xarray as xr

from oceantide.tide import Tide
from oceantide.core.utils import set_attributes


def read_oceantide(filename=None, engine=None, chunks={}, **kwargs):
    """Read oceantide file format.

    The oceantide format has complex constituents variables split into real and imag
        variables allowing for better packing and making it easier to support netcdf.

    Args:
        - filename (str): Name of Oceantide dataset to read.
        - engine (str): Engine to pass to xr.open_dataset, guessed by xarray based on
          filename extension if not specified.
        - chunks (dict): Chunk sizes along each dimensions to pass to xr.open_dataset,
          by default use chunking specified on disk.
        - kwargs: Extra kwargs to pass to xr.open_dataset.

    Returns:
        - Formatted dataset with complex variables and the tide accessor.

    """
    dset = xr.open_dataset(filename, engine=engine, chunks=chunks, **kwargs)
    dsout = dset[["dep"]]
    for v in ["h", "u", "v"]:
        dsout[v] = dset[f"{v}_real"] + 1j * dset[f"{v}_imag"]
    set_attributes(dsout, "dataset")
    dsout["con"] = dsout.con.astype("U4")
    return dsout



"""Tide xarray accessor."""
import os
import re
import glob
import types
from pathlib import Path
from importlib import import_module
from inspect import getmembers, isfunction
import datetime
import warnings
import numpy as np
import dask.array as da
import pandas as pd
import xarray as xr

from oceantide.core.utils import nodal, set_attributes
from oceantide.constituents import OMEGA


HERE = Path(__file__).parent


class Plugin(type):
    """Add output functions as bound methods at class creation."""

    def __new__(cls, name, bases, dct):
        module_names = [
            f.replace(".py", "")
            for f in glob.glob1(os.path.join(HERE, "output"), "*.py")
        ]
        modules = [import_module(f"oceantide.output.{name}") for name in module_names]
        for module_name, module in zip(module_names, modules):
            for func_name, func in getmembers(module, isfunction):
                if func_name == f"to_{module_name}":
                    dct[func_name] = func
        return type.__new__(cls, name, bases, dct)


@xr.register_dataset_accessor("tide")
class Tide(metaclass=Plugin):
    """Xarray accessor object to process tide constituents dataset."""

    def __init__(self, xarray_obj):
        self._obj = xarray_obj
        self._validate()

    def __repr__(self):
        return re.sub(r"<.+>", f"<{self.__class__.__name__}>", str(self._obj))

    def _set_attributes_output(self, dset):
        """Set attributes in output timeseries dataset."""
        set_attributes(dset, "timeseries")
        dset.attrs = {"description": "Tide elevation and currents time series"}
        return dset

    def _validate(self):
        """Check if dataset object has already been constructed.

        The following are checked to ensure it has appropriate format:
            * Required complex variables are present and share same coordinates.
            * Constituents coordinate appropriately formatted.

        """
        required_vars = ["h", "u", "v"]
        if not set(required_vars).issubset(self._obj.data_vars):
            warnings.warn(
                f"Tide accessor requires variables {required_vars} in dataset for full"
                f" functionality but only found {list(self._obj.data_vars.keys())}."
            )

        for v in required_vars:
            if v in self._obj.data_vars and not np.iscomplexobj(self._obj[v]):
                raise ValueError(f"Variable {v} must be complex type.")

        vars_dims = [self._obj[v].dims for v in required_vars]
        if len(set(vars_dims)) != 1:
            raise ValueError(f"Variables {required_vars} must share a common grid.")

        if self._obj.con.dtype.kind != "U":
            raise ValueError(
                f"`con` must be Unicode string dtype, found {self._obj.con.dtype}"
            )

    def amplitude(self, component="h"):
        """Tidal amplitude.

        :math:`\\lambda=\\sqrt{\\Re(z)+\\Im(z)}`

        Args:
            - component (str): Tidal component to calculate amplitude from,
              one of 'h', 'u', 'v'.

        Returns:
            - amp (DataArray): Amplitudes for component :math:`\\lambda(con,lat,lon)`.

        """
        darr = np.absolute(self._obj[component])
        darr.name = f"amp{component}"
        return darr

    def phase(self, component="h"):
        """Tidal phase relative to GMT.

        Args:
            - component (str): Tidal component to calculate amplitude from,
              one of 'h', 'u', 'v'.

        Returns:
            - phi (DataArray): Phases for component :math:`\\phi(con,lat,lon)`.

        """
        darr = self._obj[component]
        darr = np.rad2deg(np.arctan2(-darr.imag, darr.real)) % 360
        darr.name = f"phi{component}"
        set_attributes(darr, "dataset")
        return darr

    def predict(self, times, time_chunk=50, components=["h", "u", "v"]):
        """Predict tide timeseries.

        Args:
            times (arr): Array of datetime objects or DataArray of times to predict tide over. If an array, a new times dimension will be created.
            time_chunk (float): Time chunk size so that computation fit into memory.
            components (list): Tide variables to predict.

        Returns:
            ds (Dataset): Predicted tide timeseries components :math:`\\eta(time,lat,lon)`.

        """
        if not components or {"h", "u", "v"} - set(components) == {"h", "u", "v"}:
            raise ValueError("Choose at least one tide component h, u or v to predict")

        if isinstance(times, datetime.datetime):
            times = [times]
        if isinstance(times, (list, tuple, np.ndarray, pd.DatetimeIndex)):
            seconds_array = pd.array(times).view(int) / 1e9 - 694224000
            tsec = xr.DataArray(
                data=seconds_array,
                coords={"time": times},
                dims=("time",),
            ).chunk({"time": time_chunk})
        elif isinstance(times, xr.DataArray):
            tsec = times.astype(int) / 1e9 - 694224000
        else:
            raise TypeError(
                "times argument must be a list of datetimes, pandas.DatetimeIndex, "
                "numpy array of datetime64 or xarray DataArray of datetime64"
            )

        cons = list(self._obj.con.values)
        pu, pf, v0u = nodal(tsec[0] / 86400 + 48622.0, cons)

        # Variables for calculations
        pf = xr.DataArray(pf, coords={"con": cons}, dims=("con",))
        pu = xr.DataArray(pu, coords={"con": cons}, dims=("con",))
        v0u = xr.DataArray(v0u, coords={"con": cons}, dims=("con",))
        omega = xr.DataArray(
            data=[OMEGA[c] for c in cons], coords={"con": cons}, dims=("con",)
        )

        cos = da.cos(tsec * omega + v0u + pu)
        sin = da.sin(tsec * omega + v0u + pu)

        dset = xr.Dataset()
        if "dep" in self._obj.data_vars:
            dset["dep"] = self._obj.dep
        if "h" in components:
            dset["h"] = cos * pf * self._obj["h"].real - sin * pf * self._obj["h"].imag
        if "u" in components:
            dset["u"] = cos * pf * self._obj["u"].real - sin * pf * self._obj["u"].imag
        if "v" in components:
            dset["v"] = cos * pf * self._obj["v"].real - sin * pf * self._obj["v"].imag
        dset = dset.sum(dim="con", skipna=False)
        dset = self._set_attributes_output(dset)

        return dset
